use io;
use random;

/* Read instance data */
function input() {
    local usage = "Usage: localsolver portfolio.lsp inFileName=instanceFile"
            + " [outFileName=outputFile] [lsTimeLimit=timeLimit]";

    if (inFileName == nil) throw usage;

    inFile = io.openRead(inFileName);

    // Expected profit, in percentage of the portfolio
    expectedProfit = inFile.readDouble();

    // Number of stocks
    nbStocks = inFile.readInt();

    // Covariance among the stocks
    sigmaStocks[s in 0...nbStocks][t in 0...nbStocks] = inFile.readDouble();

    // Variation of the price of each stock
    deltaStock[s in 0...nbStocks] = inFile.readDouble();

    inFile.close();
}

/* Declare the optimization model */
function model() {
    // Proportion of the portfolio invested in each stock
    portfolioStock[s in 0...nbStocks] <- float(0, 1);

    // Risk of the portfolio
    risk <- sum[s in 0...nbStocks][t in 0...nbStocks](portfolioStock[s] *
        portfolioStock[t] * sigmaStocks[s][t]);

    // Return of the portfolio in percentage
    profit <- sum[s in 0...nbStocks](portfolioStock[s] * deltaStock[s]);

    // All the portfolio is used
    constraint sum[s in 0...nbStocks](portfolioStock[s]) == 1.0;

    // The profit is at least the expected profit
    constraint profit >= expectedProfit;

    // Minimize the risk
    minimize risk;

}

/* Parameterize the solver */
function param() {
    if (lsTimeLimit == nil) lsTimeLimit = 60;
}

/* Write the solution in a file with the following format:
 *  - for each stock, the proportion of the porfolio invested
    - the final profit in percentage of the portfolio */
function output() {
    if (outFileName != nil) {
        outFile = io.openWrite(outFileName);
        println("Solution written in file ", outFileName);
        for [s in 0...nbStocks] {
            local proportion = portfolioStock[s].value;
            outFile.println("Stock ", s, ": ", round(proportion * 1000) / 10, "%");
        }
        outFile.println("Profit: " + profit.value + "%");
    }
}
